@@ -419,15 +419,27 @@ Divmod.Runtime.Platform.methods(
         return localName;
     },
 
-//     function setAttribute(self, node, localName, namespaceURI, namespaceIdentifier, value) {
-//         if (namespaceURI !== undefined && node.setAttributeNS) {
-//             node.setAttributeNS(namespaceURI, localName, value);
-//         } else if (namespaceIdentifier !== undefined) {
-//             node.setAttribute(namespaceIdentifier + ':' + localName, value);
-//         } else {
-//             node.setAttribute(self._mangleAttributeName(localName), value);
-//         }
-//     },
+    /**
+     * Reliably set the value for a node attribute.
+     */
+    function setAttribute(self, node, localName, value, namespaceURI, namespaceIdentifier) {
+        if (namespaceURI === undefined && namespaceIdentifier === undefined) {
+            localName = self._mangleAttributeName(localName);
+        }
+        if (node.hasAttributeNS) {
+            if (node.hasAttributeNS(namespaceURI, localName)) {
+                return node.setAttributeNS(namespaceURI, localName, value);
+            } else if (node.hasAttributeNS(namespaceIdentifier, localName)) {
+                return node.setAttributeNS(namespaceIdentifier, localName, value);
+            }
+        }
+        if (node.setAttribute) {
+            var a = (namespaceIdentifier === undefined) ?
+                localName : namespaceIdentifier + ':' + localName;
+            return node.setAttribute(a, value);
+        }
+        return null;
+    },
 
     /**
      * This is _the_way_ to get the value of an attribute off of node
@@ -632,6 +644,23 @@ Divmod.Runtime.Platform.methods(
         return d;
     },
 
+    /**
+     * Load the stylesheet at the given URL by appending a C{link} tag to the
+     * document's head.
+     *
+     * @type stylesheetURL: C{String}
+     *
+     * @rtype: C{undefined}
+     */
+    function loadStylesheet(self, location) {
+        var linkNode = document.createElement('link');
+        linkNode.setAttribute('rel', 'stylesheet');
+        linkNode.setAttribute('type', 'text/css');
+        linkNode.setAttribute('href', location);
+        var headNode = document.getElementsByTagName('head')[0];
+        headNode.appendChild(linkNode);
+    },
+
     /**
      * Simulate the functionality of the DOM Level 2 Document.importNode
      * method, used on the browser document.
@@ -668,14 +697,42 @@ Divmod.Runtime.Platform.methods(
      * event, but in a delayed call scheduled from that event.
      */
     function addLoadEvent(self, callable) {
-        Divmod.Base.addLoadEvent(function() {
-                setTimeout(callable, self._loadEventDelay);
-            });
+        var func = function() { setTimeout(callable, self._loadEventDelay); };
+        Divmod.Base.addToCallStack(window, "onload", func, true);
+    },
+
+    /**
+     * Add a beforeunload event handler.
+     *
+     * @param aWindow: The window object.
+     * @param handler: The handler.
+     */
+    function addBeforeUnloadHandler(self, aWindow, handler) {
+        Divmod.Base.addToCallStack(aWindow, 'onbeforeunload', handler);
     });
 
 
+/**
+ * Spidermonkey runtime for unit testing.
+ *
+ * @type loadEvents: Array
+ * @ivar loadEvents: A list of load events that have been added.
+ */
 Divmod.Runtime.Spidermonkey = Divmod.Runtime.Platform.subclass(
     'Divmod.Runtime.Spidermonkey');
+Divmod.Runtime.Spidermonkey.methods(
+    function __init__(self, name) {
+        Divmod.Runtime.Spidermonkey.upcall(self, '__init__', name);
+        self.loadEvents = [];
+    },
+
+    /**
+     * Add the given function to loadEvents.
+     */
+    function addLoadEvent(self, callable) {
+        self.loadEvents.push(callable);
+    });
+
 Divmod.Runtime.Spidermonkey.isThisTheOne = function isSpidermonkeyTheOne() {
     try {
         window;
@@ -1141,6 +1198,13 @@ Divmod.Runtime.InternetExplorer.methods(
         }
 
         return foundNode;
+    },
+
+    /**
+     * Add the handler to the <body> element instead.
+     */
+    function addBeforeUnloadHandler(self, aWindow, handler) {
+        Divmod.Base.addToCallStack(aWindow.document.body, 'onbeforeunload', handler);
     });
 
 
@@ -1222,5 +1286,19 @@ Divmod.Runtime.Platform.determinePlatform = function determinePlatform() {
     throw new Error("Unsupported platform");
 };
 
-Divmod.Runtime.theRuntimeType = Divmod.Runtime.Platform.determinePlatform();
-Divmod.Runtime.theRuntime = new Divmod.Runtime.theRuntimeType;
+/**
+ * Initialise the Divmod runtime.
+ */
+Divmod.Runtime.initRuntime = function initRuntime(runtimeType) {
+    // If the runtime type has already been determined, then we do not attempt
+    // detection again. The platform type shouldn't be changing at runtime, and
+    // the detection heuristics can be broken by code that executes later.
+
+    if (Divmod.Runtime.theRuntimeType === undefined) {
+        Divmod.Runtime.theRuntimeType = Divmod.Runtime.Platform.determinePlatform();
+    }
+
+    Divmod.Runtime.theRuntime = new Divmod.Runtime.theRuntimeType;
+}
+
+Divmod.Runtime.initRuntime();
