@@ -3,7 +3,7 @@
 """
 Generate an unhandled SIGSEGV for this process immediately upon import.
 
-@see L{nevow.test.test_testutil.JavaScriptTests.test_signalledExit}.
+@see: L{nevow.test.test_testutil.JavaScriptTests.test_signalledExit}.
 """
 
 import os, signal
